/*

Copyright (C) 2006 Christian Kreibich <christian@whoop.org>.

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to
deal in the Software without restriction, including without limitation the
rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
sell copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies of the Software and its documentation and acknowledgment shall be
given in the documentation and software packages that this Software was
used.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
THE AUTHORS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

*/
#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include <sys/types.h>
#include <sys/wait.h>
#include <sys/socket.h>
#include <sys/stat.h>
#include <regex.h>
#include <unistd.h>
#include <fcntl.h>
#include <signal.h>
#include <errno.h>

#include <libnd.h>
#include <plugins/libnd_traceset.h>
#include <plugins/libnd_mux.h>

typedef struct libnd_sort {

  gboolean            do_check;
  gboolean            keep_temps;
  
  char               *output;
  struct bpf_timeval  last;
  int                 count;

  pcap_t             *pcap;
  pcap_dumper_t      *dumper;
  char               *dumper_name;

  LND_Mux            *muxer;
} LND_Sort;


static void
sort_help(void)
{
  printf("This plugin sorts packets in pcap traces chronologically..\n"
	 "\n"
	 "USAGE: lndtool -r sort --output|-o <FILE> [--help|h] [--check|-c] [--names-file|-f <FILE>] [<TRACE2> ...]\n"
	 "\n"
	 "  --help, -h, -?           This message.\n"
	 "  --check|-c               Prints out timestamp of every packet that violates order.\n"
	 "  --keep-temps|-k          Keep temporary files. For debugging.\n"
	 "  --output|-o <FILE>       Name of resulting trace.\n"
	 "  --names-file|-f <FILE>   Take filenames of traces from flatfile FILE.\n");
}

static LND_Sort *
sort_new(void)
{
  LND_Sort *sort = g_new0(LND_Sort, 1);

  if (! (sort->muxer = libnd_mux_new()))
    return NULL;

  if (! (sort->dumper_name = malloc(sizeof(char) * MAXPATHLEN)))
    {
      libnd_mux_free(sort->muxer);
      g_free(sort);
      return NULL;
    }

  return sort;
}

static void
sort_free(LND_Sort *sort)
{  
  libnd_mux_free(sort->muxer);
  g_free(sort->dumper_name);
  g_free(sort);
}

static gboolean
sort_advance_dumper(LND_Sort *sort, gboolean create)
{
  if (sort->dumper)
    {
      pcap_dump_close(sort->dumper);

      /* Must come *after* close, so that a check for valid
       * pcap file format succeeds!
       */
      libnd_traceset_add_trace_name(sort->muxer->traces, sort->dumper_name);
    }
  
  if (create)
    {
      g_snprintf(sort->dumper_name, MAXPATHLEN, "__lnd_sort_tmp_.%i", sort->count++);
      
      if (! (sort->dumper = pcap_dump_open(sort->pcap, sort->dumper_name)))
	{
	  fprintf(stderr, "Cannot open dump: %s\n", pcap_geterr(sort->pcap));
	  return FALSE;
	}
    }
  
  return TRUE;
}

static gboolean
sort_trace(LND_TraceSet *set, LND_Trace *trace, LND_Sort *sort)
{
  LND_PacketIterator  pit;
  LND_Packet *packet;

  sort->pcap = pcapnav_pcap(trace->tpm->base->pcn);

  if (! sort->do_check)
    sort_advance_dumper(sort, TRUE);

  /* Iterate over entire individual trace, read-only. */
  for (libnd_pit_init_mode(&pit, trace, LND_PACKET_IT_AREA_R);
       libnd_pit_get(&pit); libnd_pit_next(&pit))
    {
      packet = libnd_pit_get(&pit);

      if (sort->do_check)
	{
	  if (pcapnav_timeval_cmp(&sort->last, &packet->ph.ts) > 0)
	    printf("%lu.%lu\n", (long unsigned) packet->ph.ts.tv_sec,
		   (long unsigned) packet->ph.ts.tv_usec);
	}
      else
	{
	  if (pcapnav_timeval_cmp(&sort->last, &packet->ph.ts) > 0)
	    sort_advance_dumper(sort, TRUE);
	  
	  pcap_dump((u_char*) sort->dumper, &packet->ph, packet->data);	  
	}

      sort->last = packet->ph.ts;
    }
  
  return TRUE;
  TOUCH(set);
}

static gboolean
sort_temp_unlink_cb(LND_TraceSet *set, const char *name, void *unused)
{
  unlink(name);
  return TRUE;

  TOUCH(set);
  TOUCH(unused);
}

/* ---- Plugin symbols below ------------------------------------------- */

const char *
name(void)
{
  return "Sort";
}

const char *
author(void)
{
  return "Christian Kreibich, <christian@whoop.org>";
}

const char *
version(void)
{
  return "0.1";
}


gboolean
init(void)
{
  if (! libnd_plugin_find("Trace-Set"))
    {
      D(("Trace-Set plugin not found\n"));
      return FALSE;
    }

  if (! libnd_plugin_find("Mux"))
    {
      D(("Mux plugin not found\n"));
      return FALSE;
    }

  D(("Sort plugin initialization complete.\n"));
  
  return TRUE;
}

gboolean
run(LND_Trace *trace, LND_PluginArgs *args)
{
  LND_Sort *sort;
  LND_TraceSet *traces;
  int i, files = 0;

  if (! (traces = libnd_traceset_new()))
    return FALSE;

  if (! (sort = sort_new()))
    {
      libnd_traceset_free(traces);
      return FALSE;
    }

  for (i = 0; i < args->argc; i++)
    {
      if (!strcmp(args->argv[i], "-h")     ||
	  !strcmp(args->argv[i], "--help") ||
	  !strcmp(args->argv[i], "-?"))
	{
	  sort_help();
	  return FALSE;
	}
      else if (!strcmp(args->argv[i], "-f") ||
	       !strcmp(args->argv[i], "--names-file"))
	{
	  if (++i == args->argc)
	    {
	      printf("You need to pass the name of a file that contains\n"
		     "a list of trace files, one filename per line, when\n"
		     "using the --names-file|-f option.\n");
	      return FALSE;
	    }
	  
	  if (! libnd_misc_can_read(args->argv[i]))
	    printf("Could not open '%s'\n", args->argv[i]);
	  else if (! libnd_traceset_add_trace_name_list(traces, args->argv[i]))
	    printf("Could not add all traces provided in '%s'\n", args->argv[i]);
	  else
	    files++;
	}
      else if (!strcmp(args->argv[i], "-o") ||
	       !strcmp(args->argv[i], "--output"))
	{
	  if (++i == args->argc)
	    {
	      printf("You need to pass the name of the resulting trace\n"
		     "to --output|-o.\n");
	      return FALSE;
	    }
	  
	  g_free(sort->output);
	  sort->output = g_strdup(args->argv[i]);
	}
      else if (!strcmp(args->argv[i], "-c") || !strcmp(args->argv[i], "--check"))
	{
	  sort->do_check = TRUE;
	}
      else if (!strcmp(args->argv[i], "-k") || !strcmp(args->argv[i], "--keep-temps"))
	{
	  sort->keep_temps = TRUE;
	}
      else if (args->argv[i][0] != '-')
	{
	  if (! libnd_traceset_add_trace_name(traces, args->argv[i]))
	    printf("Trace file %s not added.\n", args->argv[i]);
	  else
	    files++;
	}
      else
	printf("Warning: option %s unknown.\n", args->argv[i]);
    }
  
  if (args->argc == 0)
    {
      sort_help();
      return FALSE;
    }

  if (! sort->do_check && ! sort->output)
    {
      printf("You need to provide an output trace name via --output|-o.\n");
      return FALSE;
    }

  /* Create sorted output file */
  libnd_traceset_foreach(traces, (LND_TraceSetCB) sort_trace, sort);

  if (! sort->do_check)
    {
      if (sort->dumper)
	sort_advance_dumper(sort, FALSE);
      
      /* Now merge them */
      libnd_mux_set_output(sort->muxer, sort->output);
      libnd_mux(sort->muxer);
    }
  
  /* Cleanup temp file */
  if (! sort->keep_temps)
    {
      libnd_traceset_foreach_name(sort->muxer->traces,
				  sort_temp_unlink_cb, NULL);
    }

  sort_free(sort);
  libnd_traceset_free(traces);

  return TRUE;
  TOUCH(trace);
}
